from __future__ import print_function

import numpy as numpy
import math as math
import sys
import trace

__author__ = "Do Kester"
__year__ = 2018
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2016 - 2017 Do Kester


def setNoneAttributes( obj, name, value, listNone ) :
    """
    Set attribute contained in dictionary dictList into the attr-list.
    A list is a native list or a numpy.ndarray. It also checks the type.
    if values is a singular item of the proper type it will be inserted as [value].

    Parameters
    ----------
    obj : object
        to place the attribute in
    name : str
        of the attribute
    value : any
        of the attribute
    listNone : list of names
        that could have a None value

    Returns
    -------
        True on succesful insertion. False otherwise.

    Raises
    ------
        TypeError   if the type is not as in the dictionary
     """
    if name in listNone and value is None :
        object.__setattr__( obj, name, value )
        return True
    return False


def setListOfAttributes( obj, name, value, dictList ) :
    """
    Set attribute contained in dictionary dictList into the attr-list.
    A list is a native list or a numpy.ndarray. It also checks the type.
    if values is a singular item of the proper type it will be inserted as [value].

    Parameters
    ----------
    obj : object
        to place the attribute in
    name : str
        of the attribute
    value : any
        of the attribute
    dictList : dictionary
        of possible attributes {"name": type}

    Returns
    -------
        True on succesful insertion. False otherwise.

    Raises
    ------
        TypeError   if the type is not as in the dictionary

    """
    if name in dictList :
        _type = dictList[name]
        isl = isList( value, _type )
#        print( name, isl )
        if ( isl[0] ) :
            if _type is not int and _type is not float :
                _array = value if isl[1] else [value]
            elif ( isl[1] ) : _array = numpy.asarray( value, dtype=_type )
            else : _array = numpy.asarray( [value], dtype=_type )
            object.__setattr__( obj, name, _array )
            return True
        else :
            TypeError( name + ' has not the proper type: (list of) ' + str( dictList[name] ) )
    return False


def setSingleAttributes( obj, name, value, dictSingle ) :
    """
    Set a singular attribute contained in dictionary dictSingle into the attr-list.
    It also checks the type.

    Parameters
    ----------
    obj : object
        to place the attribute in
    name : str
        of the attribute
    value : any
        of the attribute
    dictSingle : dictionary
        of possible attributes {"name": type}

    Returns
    -------
        True on succesful insertion. False otherwise.

    Raises
    ------
        TypeError   if the type is not as in the dictionary
     """
    if name in dictSingle :
        if isInstance( value, dictSingle[name] ) :
            object.__setattr__( obj, name, value )
            return True
        else :
            TypeError( name + ' has not the proper type: ' + str( dictSingle[name] ) )
    return False

def makeNext( x, k ) :
    """
    Return next item of x, and last item if x is exhausted.
    Or x itself if x is singular.
    """
    try :
        _xnext = x[-1]
    except :
        _xnext = x
    while True :
        try :
            _xnext = x[k]
            yield _xnext
            k += 1
        except :
            yield _xnext


def length( x ) :
    """
    Return the length of any item. Singletons too.
    """
    try :
        return len( x )
    except :
        return 1

def toArray( x, ndim=1, dtype=None ) :
    """
    Return a array of x when x is a number

    Parameters
    ----------
    x : any number, list/array of numbers or []
        to be converted to numpy.ndarray
    ndim : int
        minimum number of dimensions present
    dtype : type
        conversion to type (None : as is)

    """
    return numpy.array( x, dtype=dtype, copy=False, ndmin=ndim )

def isList( item, cls ) :
    """
    Return (True,False) if item is a instance of cls
           (True,True)  if item is a (list|ndarray) of instances of cls
           (False,False) if not
    """
#    print( item, cls, item.__class__ )
    if isInstance( item, cls ) : return (True,False)
    islst = isinstance( item, list ) or isinstance( item, numpy.ndarray )
    if islst :
        for i in numpy.asarray( item ).flat :
            islst = islst and isInstance( i, cls )
    return (islst,islst)

def isInstance( item, cls ) :
    """
    Returns true when one of the following is true
    1. when cls is int   : item is an int or item is a numpy.integer.
    2. when cls is float : item is an float or item is an int.
    3. when cls is cls   : item is a cls.
    """
    if cls is int :
        return isinstance( item, int ) or isinstance( item, numpy.integer )
    elif cls is float :
        return isinstance( item, float ) or isInstance( item, int )
    else :
        return isinstance( item, cls )

def ndprint( x, form='{0:.3f}' ) :
    """
    Print a ndarray, formatted.
    """
    print( numpy.array2string(x, formatter={'float_kind':form.format}) )

def decorate( src, des, copy=True ) :
    """
    Transfer attributes from src to des.
    If copy is True try to copy the attributes, otherwise link it.

    Parameters
    ----------
    src : object
        source of the attributes
    des : object
        destiny for the attributes
    copy : bool
        if True: copy
    """
    atr = vars( src )
    ld = list( atr.keys() )
    for key in ld :
        value = atr[key]
        if copy :
            try :
                value = value.copy()
            except :
                pass
        object.__setattr__( des, key, value )


def printclass( cls, nitems=8 ) :
    """
    Print the attributes of a class.
    """
    print( "+++++++++++++++++++++++++++++++++++++++++++++++++++++++" )
    print( cls )
    print( "+++++++++++++++++++++++++++++++++++++++++++++++++++++++" )
    atr = vars( cls )
    ld = list( atr.keys() )
    ld.sort()
    for key in ld :
        print( "%-16.16s"%key, end="" )
        val = atr[key]
        nv = length( val )
        if isinstance( val, (list,numpy.ndarray) ) :
            print( "[", end="" )
            for k in range( min( nitems, nv ) ) :
                print( val[k], " ", end="" )
            if nv > nitems : print( "... ", end="" )
            print( "]" )
        elif key == "model" :
            print( val.shortName( ) )
        else :
            print( val )

def nicenumber( x ) :
    """
    Return a nice number close to x.
    """
    if x == 0 :
        return x
    n = 1.0
    while x > 10 :
        x /= 10
        n *= 10
    while x < 1 :
        x *= 10
        n /= 10
    return int( x ) * n

def fix2int( x ) :
    """
    Return integer array with values as in x
    Parameters
    ----------
    x : array_like
        array of integer floats
    """
    return numpy.fix( x + 0.000001 ).astype( int )

def track( statement ) :
    """

    Parameters
    ----------
    statement : str
        statement to be traced

    """
    # create a Trace object, telling it what to ignore, and whether to
    # do tracing or line-counting or both.
    tracer = trace.Trace( ignoredirs=[sys.prefix, sys.exec_prefix], trace=0, count=1 )

    # run the new command using the given tracer
    tracer.run( statement )

    # make a report, placing output in the current directory
    r = tracer.results()
    r.write_results(show_missing=True, coverdir=".")



# run with : python3 -m unittest TestKernels

import unittest
import numpy as numpy
from astropy import units
import matplotlib.pyplot as plt
import warnings

from BayesicFitting import *

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestKernels( unittest.TestCase ):
    """
    Test harness for Models

    Author:      Do Kester

    """
    def plotKernels( self ) :
        self.testKernels( plot=True )

    def plotHuber( self ) :
        kernel = Huber()
        xhm = kernel.fwhm / 2
        self.assertAlmostEqual( kernel.result( xhm ), 0.5 )
        self.assertAlmostEqual( kernel.result( -xhm ), 0.5 )
        self.plotK( [Huber] )



    def testKernels( self, plot=False ):
        kernels = [Gauss, Lorentz, Sinc, Biweight, Cosine, CosSquare, Parabola, Triangle,
                   Tricube, Triweight, Uniform]
        if plot :
            self.plotK( kernels )

        for kernl in kernels :
            kernel = kernl()
            self.stdKerneltest( kernel, plot=plot )
            self.assertTrue( kernel.isBound() or ( kernl in kernels[:3] ) )


    def stdKerneltest( self, kernel, plot=None ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        x += 0.02

        model = KernelModel( kernel=kernel )
        print( "********************************************" )
        print( model )

        numpy.set_printoptions( precision=3, suppress=False )

        par = [1.0,0.0,1.0]
        print( model.partial( x, par ) )
        model.testPartial( x[5], model.parameters )

        model.testPartial( x[0], par )
        model.xUnit = units.m
        model.yUnit = units.kg
        for k in range( model.getNumberOfParameters() ):
            print( "%d  %-12s  %-12s"%(k, model.getParameterName( k ), model.getParameterUnit( k ) ) )

        xx = numpy.linspace( 0, 101, 101000 )
        yy = model.result( xx )
        sn = numpy.sum( yy ) / 500
        ss = kernel.integral

        print( "Integral   ", sn, model.getIntegralUnit( ) )
        self.assertAlmostEqual( sn, 1.0, 1 )
        self.assertAlmostEqual( ss, 1.0 / model.parameters[0], 4 )

        xhm = kernel.fwhm / 2
        print( xhm, kernel.result( xhm ), kernel.result( -xhm ) )

        self.assertAlmostEqual( kernel.result( xhm ), 0.5 )
        self.assertAlmostEqual( kernel.result( -xhm ), 0.5 )

        part = model.partial( x, par )
        nump = model.numPartial( x, par )
        print( part.shape, nump.shape )
        for k in range( 11 ) :
            print( "%3d %8.3f"%(k, x[k]), part[k,:], nump[k,:] )

        k = 0
        np = model.npbase
        for (pp,nn) in zip( part.flatten(), nump.flatten() ) :
            print( "%3d %10.4f  %10.4f %10.4f"%(k, x[k//np], pp, nn) )
            self.assertAlmostEqual( pp, nn, 3 )
            k += 1

        mc = model.copy( )
        for k in range( mc.getNumberOfParameters() ):
            print( "%d  %-12s  %-12s"%(k, mc.getParameterName( k ), mc.getParameterUnit( k ) ) )
        mc.parameters = par
        model.parameters = par

        for (k,xk,r1,r2) in zip( range( x.size ), x, model.result( x ), mc.result( x ) ) :
            print( "%3d %8.3f  %10.3f %10.3f" % ( k, xk, r1, r2 ) )
            self.assertEqual( r1, r2 )

    def plotK( self, kernels ) :
        x = numpy.linspace( -5, +5, 1001 )
        par = [1.0,0.0,1.0]
        for kernel in kernels :
            model = kernel()
            print( model )
            y = model.result( x )
            plt.plot( x, y, '-', linewidth=2 )
            x2 = numpy.linspace( -4, +4, 11 )
            dy = model.partial( x2 )
            print( dy )
            yy = model.result( x2 )
            print( yy )
            for k in range( 11 ) :
                x3 = numpy.asarray( [-0.05, +0.05] )
                y3 = x3 * dy[k] + yy[k]
                plt.plot( x2[k] + x3, y3, 'r-' )

        plt.show()


    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestModels.__class__ )

if __name__ == '__main__':
    unittest.main( )





import numpy as numpy
from .Model import Model

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (JAVA code)
#  *    2016 - 2017 Do Kester

class LinearModel( Model ):
    """
    Anchestor of all linear models.

    LinearModel implements the baseResult method needed in all linear models.

    For Linear models it holds that
    .. math::
        f( x:p ) = \sum( p_i * f^\prime( x ) )

    which means that only the partial derivatives to :math:`p_i`,
    the :math:`f^\prime( x )`, need to be given as ``basePartial``.
    The ``baseResult`` follows from that one.
    It is implemented here.


    """
    def __init__( self, nparams, ndim=1, copy=None, **kwargs ):
        """
        class for all linear models.

        Parameters
        ----------
        nparams : int
            the number of parameters in this model
        ndim : int
            the dimensionality of the inputs (default: 1)
        copy : LinearModel
            model to be copied (default: None)

        """
        super( LinearModel, self ).__init__( nparams, ndim=ndim, copy=copy, **kwargs )

    def baseResult( self, xdata, params ):
        """
        Returns the base result of linear models.

        for linear models the result is the inner product of parameters
        and partial derivatives.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        parlist = numpy.arange( self.npmax )
        part = self.basePartial( xdata, params, parlist=parlist  )

#        print( "LM  ", params )
#        print( "LM  ", part )
#        print( "LM  ", parlist )
#        try :
#            return numpy.inner( params, part )
#        except ValueError :
#            print( "LinearModel: ValueError" )
#            pass

        res = numpy.zeros( part.shape[0], dtype=float )
        for k in parlist :
            res += params[k] * part[:,k]


#        for k,p in enumerate( params ) :
#            res += p * part[:,k]
        return res




import numpy as numpy
from astropy import units
import math
from . import Tools
from .Formatter import formatter as fmt

from .Engine import Engine
from .Engine import DummyPlotter

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester

class GalileanEngine( Engine ):
    """
    Move all parameters in forward steps, with optional mirroring on the edge.

    Move the parameters in a random direction for N iterations; mirror the direction
    on the gradient of the logLikelihood when the parameters enter the zone of logLlow.

    Attributes
    ----------
    walkers : SampleList
        walkers to be diffused
    errdis : ErrorDistribution
        error distribution to be used
    nstep : int (10)
        average number of steps to be taken
    size : float (0.5)
        average normalized stepsize
    maxtrials : int
        maximum number of trials for various operations
    rng : numpy.random.RandomState
        random number generator


    Author       Do Kester.

    """

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, copy=None, seed=4213 ):
        """
        Default Constructor.

        Parameters
        ----------
        walkers : SampleList
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        copy : GalileanEngine
            to be copied
        seed : int
            for random number generator
        """
        super( GalileanEngine, self ).__init__( walkers, errdis, copy=copy,
                        seed=seed  )
        self.nstep = 3
        self.size = 0.5

        self.plotter = DummyPlotter( )

    def copy( self ):
        """ Return copy of this.  """
        engine = GalileanEngine( self.walkers, self.errdis, copy=self )
        engine.nstep = self.nstep
        engine.size = self.size
        return engine

    def __str__( self ):
        """ Return the name of this engine.  """
        return str( "GalileanEngine" )

    #  *********EXECUTE***************************************************
    def execute( self, walker, lowLhood, fitIndex=None ):
        """
        Execute the engine by diffusing the parameters.

        Parameters
        ----------
        walker : Sample
            walker to diffuse
        lowLhood : float
            lower limit in logLikelihood
        fitIndex : array_like
            list of the/some parameters indices to be diffused

        Returns
        -------
        int : the number of successfull moves

        """
        model = walker.model
        Lhood = walker.logL
        allpars = walker.allpars
        fitIndex = walker.fitIndex

#        reportCall( )

        npout = 0
        inside = 0
        Ltry = 0

        self.plotter.start( )
#        print( "GE  UM  ", walker.id, allpars, fitIndex )
        um = UnitMovements( model, allpars, fitIndex, self )

        ptry = allpars.copy()
#        print( ptry )
#        print( um.upar )
#        print( self.unit2Domain( model, um.upar ) )

        nstep = int( self.nstep * ( 1 + self.rng.rand() ) )
        maxtrial = self.maxtrials * nstep
        step = 0
        trial = 0

        while True:
            trial += 1
            um.setParameters( model, allpars )
            if inside == 0 :                            # safely inside lowLhood area
                ptry[fitIndex] = um.stepPars( 1.0 )

                self.plotter.move( allpars, ptry, 0 )
            elif inside == 1 :                          # first time outside -> mirror
                f = ( Lhood - lowLhood ) / ( Lhood - Ltry )     # lin interpolation to edge
                pedge = ptry.copy()
                pedge[fitIndex] = um.stepPars( f )                # ptry on edge
                dLdp = self.errdis.partialLogL( model, pedge, fitIndex )
                self.plotter.move( allpars, pedge, 1 )

                um.mirrorOnLowL( dLdp )
                ptry[fitIndex] = um.stepPars( 1 - f )

                self.plotter.move( pedge, ptry, 2 )
            else:                                       # mirroring failed; do reverse
                um.reverseVelocity( 1.0 )
                ptry[fitIndex] = um.stepPars( 1.0 )
                self.size *= 0.7

                self.plotter.move( allpars, ptry, 3 )

            Ltry = self.errdis.logLikelihood( model, ptry )

            if Ltry >= lowLhood:
                allpars = ptry.copy( )
                Lhood = Ltry
                self.reportSuccess( )
                npout = len( allpars )
                inside = 0
                step += 1
            else:
                inside += 1
                if inside == 1:
                    self.reportReject( )
                else:
                    self.reportFailed( )

#            print( "GEng  ", model.npchain, fitIndex, fmt(ptry), fmt(Ltry),
#                inside, step, fmt( self.size ) )

            if not ( step < nstep and trial < maxtrial ):
                break

        self.setSample( walker, model, allpars, Lhood )

        if trial >= maxtrial:
#            self.size = self.size * 0.9 + 0.00001
            pass

        self.plotter.stop()
        return npout

import matplotlib.pyplot as plt

class UnitMovements( object ):

    def __init__( self, model, allpars, fitIndex, engine ):
        self.model = model
        self.np = len( fitIndex )
        self.fitIndex = fitIndex
        self.engine = engine
        self.setParameters( model, allpars )

#        if fitIndex[-1] >= len( self.engine.unitRange ) :
#            print( "GAL  ", self.engine.unitRange, fitIndex, end="" )
#            self.engine.unitRange = numpy.append( self.engine.unitRange, [1.0] )
#            print( self.engine.unitRange )

        self.upran = self.engine.unitRange[fitIndex]
        self.setVelocity( self.engine.size )

    def setParameters( self, model, allpars ):
        self.upar = self.engine.domain2Unit( model, allpars, kpar=self.fitIndex )

    def mirrorOnLowL( self, dLdp ):
        inprod = numpy.inner( dLdp, self.uvel )
        sumsq  = numpy.inner( dLdp, dLdp )
        self.uvel -= 2 * dLdp * inprod / sumsq

    def setVelocity( self, size ):
        if self.model.isDynamic():
            self.setVelocityDynamic( size )
        else:
            self.setVelocityStatic( size )

    def allpars2unit( self, kw ) :
        allpars = self.engine.walkers[kw].allpars
        return self.engine.domain2Unit( self.model, allpars, kpar=self.fitIndex )

    def setVelocityStatic( self, size ):
        # find two randomly chosen walkers
        nm = len( self.engine.walkers )
        k1 = k0 = self.engine.rng.randint( nm )
        while k1 == k0:
            k1 = self.engine.rng.randint( nm )

        # subtract the parameter postions to get a velocity
        self.uvel = ( self.allpars2unit( k0 ) - self.allpars2unit( k1 ) ) * size

        # add a random contibution
        rv = self.uniform() * self.upran * size
        self.uvel = ( nm * self.uvel + self.np * rv ) / ( nm + self.np )

    def uniform( self ) :
        return self.engine.rng.rand( self.np ) - 0.5

    def setVelocityDynamic( self, size ):
#        print( "SVD   ", self.np, size, len( self.engine.walkers[0].allpars), self.engine.unitRange )
        self.uvel = self.uniform() * size * self.engine.unitRange[self.fitIndex]

    def reverseVelocity( self, size ):
        upv = self.uvel                 # keep original velocity
        self.setVelocity( 0.5 * size )  # get a new one
        self.uvel -= ( 0.5 * upv )      # subtract original

    def stepPars( self, f ):
        uv = self.uvel
        pv = self.upar + uv * f
        # check if outside [0,1]
        pv, uv = numpy.where( pv <= 0, ( -pv, -uv ), ( pv, uv ) )
        self.upar, self.uvel = numpy.where( pv >= 1, ( 2 - pv, -uv ), ( pv, uv ) )
        return self.engine.unit2Domain( self.model, self.upar, kpar=self.fitIndex )




import numpy as numpy
from threading import Thread, current_thread

#from OrderProblem import OrderProblem

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester


threadErrors = []

class Explorer( object ):
    """
    Explorer is a helper class of NestedSampler, which contains and runs the
    diffusion engines.

    It uses Threads to parallelise the diffusion engines.

    Attributes
    ----------
    walkers : SampleList
        samples to be explored
    engines : [engine]
        list of engines to be used
    rng : numpy.random.RandomState
        random number generator
    rate : float (1.0)
        governs processing speed (vs precision)
    maxtrials : int (5)
        number of trials
    verbose : int (0)
        level of blabbering
    lowLhood : float
        present low likelihood level
    generation : int
        counting explorer calls

    Author       Do Kester.

    """
    TWOP32 = 2 ** 32

    def __init__( self, ns ):
        """
        Construct Explorer from a NestedSampler object.
        Parameters
        ----------
        ns : NestedSampler
            the calling NestedSampler

        """
        self.walkers = ns.walkers
        self.engines = ns.engines
        self.rng = ns.rng
        self.rate = ns.rate
        self.maxtrials = ns.maxtrials
        self.verbose = ns.verbose
#        if isinstance( self.walkers[0].model, OrderProblem ) :
#           return
        self.engines[0].calculateUnitRange( )

    def explore( self, worst, lowLhood, fitindex ):
        """
        Explore the likelihood function, using threads.

        Parameters
        ----------
        worst : [int]
            list of samples to be explored/updated
        lowLhood : float
            level of the low likelihood
        fitindex : list of int
            list of parameter indices to fit

        """
        self.lowLhood = lowLhood
        explorerThreads = []
        for kw in worst :
            seed = self.rng.randint( self.TWOP32 )
#            print( "Thr  %d  "%kw, seed )
            exThread = ExplorerThread( "explorer_%d"%kw, kw, self, seed, fitindex )
            exThread.start( )
            explorerThreads += [exThread]

        for thread in explorerThreads :
            thread.join( )

            for k,e in enumerate( self.engines ) :
                nc = 0
                for i in range( 3 ) :
                    nc += thread.engines[k].report[i]
                    e.report[i] += thread.engines[k].report[i]
                e.report[3] += nc

        if len( threadErrors ) > 0: #check if there are any errors
            for e in threadErrors:
                print( e )
            raise Exception( "Thread Error" )

        # recalculate  TBC
#        if isinstance( self.walkers[0].model, OrderProblem ) :
#            return

        self.engines[0].calculateUnitRange( )

class ExplorerThread( Thread ):
    """
    One thread for the Explorer. It updates one walker.

    Attributes
    ----------
    id : int
        identity for thread
    walkers : SampleList
        list of walkers
    rng : numpy.random.RandomState
        random number generator
    engines : [Engine]
        copy of the list of Engines of Explorer
    errdis : ErrorDistribution
        to be used (=self.engines[0].errdis)
    """

    global threadErrors

    def __init__( self, name, id, explorer, seed, fitindex ):
        super( ExplorerThread, self ).__init__( name=name )
        self.id = id
        self.explorer = explorer
        self.fitindex = fitindex
        self.engines = [eng.copy() for eng in explorer.engines]
        self.rng = numpy.random.RandomState( seed )
        self.errdis = self.engines[0].errdis
        self.verbose = explorer.verbose

    def run( self ):
        try :
            self.explore( self.id, self.fitindex )
        except Exception as e :
            threadErrors.append( [repr(e) + " occurred in thread %d"%self.id] )
            raise


    def explore( self, walkerId, fitindex ):
        walker = self.explorer.walkers[walkerId]
        oldlogL = walker.logL

        maxmoves = len( fitindex ) / self.explorer.rate
        maxtrials = self.explorer.maxtrials / self.explorer.rate

        lowLhood = self.explorer.lowLhood
        moves = 0
        trials = 0
        while moves < maxmoves and trials < maxtrials :
            i = 0
            for engine in self.rng.permutation( self.engines ) :
#                print( "Exp   ", engine, walker.id, fitindex, walker.allpars, walker.fitIndex )
                moves += engine.execute( walker, lowLhood, fitindex )

                if self.verbose >= 4:
                    print( "%4d %-15.15s %4d %8.1f %8.1f ==> %3d  %8.1f"%
                            ( i, engine, walkerId, lowLhood, oldlogL, moves,
                                walker.logL ) )
                    i += 1
                    oldlogL = walker.logL

            trials += 1

        if moves == 0 :
            self.logLcheck( walker )

        return

    def logLcheck( self, walker ) :
        wlogL = self.errdis.logLikelihood( walker.model, walker.allpars )
        if wlogL != walker.logL :
            raise ValueError( "Inconsistency between stored logL %f and calculated logL %f" %
                                ( walker.logL, wlogL ) )




import numpy as numpy
from astropy import units
import math
from . import Tools
from .Prior import Prior

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2016 - 2017 Do Kester

class HyperParameter( object ):
    """
    Values and priors for the parameter(s) of an ErrorDistribution.

    Hyperparameters are not directly related to the model, they are
    parameters of the error distribution.

    Information about the scale of the noise is stored in a derived class,
    noiseScale.

    The full use of priors is reserved for Bayesian calculations as
    in NestedSampler

    Attributes
    ----------
    hypar : float
        the value of the hyperparameter.  Default: 1.0
    stdev : float
        the standard deviation of the hyperparameter.  Default: None
    prior : Prior
        the prior for the hyperparameter.
    isFixed : boolean
        keep the hyperparameter fixed at the value given by hypar.
        default: True

    """

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, hypar=1, isFixed=True, prior=None, limits=None,
                    copy=None ):
        """
        Constructor.

        Parameters
        ----------
        hypar : float
            value of the hyperparameter
        isFixed : bool
            True:   Consider the hyperparameter as fixed
            False:  Optimize the parameter too (when relevant)
                    It might need a prior and/or limits to be set
        prior : None or Prior
            None : no prior set
            prior probability on the hyperparameter
        limits : None or list of 2 floats [lo,hi]
            low limit and high limit on hypar.
        copy : HyperParameter
            HyperParameter to copy

        """
        super( HyperParameter, self ).__init__( )
        self.hypar = hypar
        self.isFixed = isFixed
        self.stdev = None
        self.prior = prior
        if prior is not None :
            self.prior.setLimits( limits )
            self.isFixed = False

        if copy is not None :
            self.hypar = copy.hypar
            self.stdev = copy.stdev
            if copy.prior is not None :
                self.prior = copy.prior.copy()
            self.isFixed = copy.isFixed

    def copy( self ):
        """ Return a copy.  """
        return HyperParameter( copy=self )

    def checkPrior( self ) :
        if self.prior is None :
            raise ValueError( "Need a prior set to %s" % self.__str__() )

    def setLimits( self, limits ):
        """
        Set the limits on the scale within the prior.

        Parameters
        ----------
        limits : list of 2 float
            the [low,high] limits.

        """
        self.checkPrior()
        self.prior.setLimits( limits )
        self.isFixed = False

    def getLimits( self ):
        """ Return the limits on the scale. """
        return self.prior.getLimits( )

    def isBound( self ) :
        """ Return true is the itergral over the prior is bound. """
        self.checkPrior()
        return self.prior.isBound()

    def domain2Unit( self, dval ):
        """
        Return a value in [0,1] given a value within the valid domain of
        a parameter for the prior distribution.

        Parameters
        ----------
        dval : float
            value within the domain of a parameter

        """
        self.checkPrior()
        return self.prior.domain2Unit( dval )

    def unit2Domain( self, uval ):
        """
        Return a value within the valid domain of the parameter given a value
        between [0,1] for the prior distribution.

        Parameters
        ----------
        uval : float
            value within [0,1]

        """
        self.checkPrior()
        return self.prior.unit2Domain( uval )

    def partialDomain2Unit( self, dval ):
        """
        Return a the derivate of the domain2Unit function to dval.

        Parameters
        ----------
        dval : float
            value within the domain of a parameter

        """
        self.checkPrior()
        return self.prior.partialDomain2Unit( dval )

    def __str__( self ) :
        return str( "HyperParameter. value = %f" % self.hypar )



import numpy as numpy
from . import Tools

from .Engine import Engine

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester

class CrossEngine( Engine ):
    """
    Cross over between 2 walkers.

    Select another walker from the ensemble and perform a cross-over of
    their parameter lists. I.e. choose at random from one list or the other.

    The walker is kept when the logLikelihood > lowLhood.

    Author       Do Kester.

    """
    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, copy=None, seed=4213 ):
        """
        Constructor.

        Parameters
        ----------
        walkers : SampleList
            walkers to be diffused
        errdis : ErrorDistribution
            error dstribution to be used
        copy : CrossEngine
            engine to copy
        seed : int
            for rng

        """
        super( CrossEngine, self ).__init__( walkers, errdis, copy=copy, seed=seed )

    def copy( self ):
        """ Return copy of this.  """
        return CrossEngine( self.walkers, self.errdis, copy=self )

    def __str__( self ):
        return str( "CrossEngine" )

    #  *********EXECUTE***************************************************
    def execute( self, walker, lowLhood, fitIndex=None ):
        """
        Execute the engine by diffusing the parameters.

        Parameters
        ----------
        walker : Sample
            walker to diffuse
        lowLhood : float
            lower limit in logLikelihood
        fitIndex : array_like
            list if the parameters to be diffused

        Returns
        -------
        int : the number of successfull moves

        """
        if fitIndex is None :
            fitIndex = walker.fitIndex
        nf = len( fitIndex )

        model = walker.model
        param = walker.allpars[:]

        nm = len( self.walkers )
        np = len( param )
        if np <= 2 and nm > 1 :
            self.reportFailed( )
            return 0

        kk = 0
        while True:
            kk += 1
            while True:
                m = self.rng.randint( nm )
                if not ( (m == walker.id or m == walker.parent) ):
                    break

            crospar = self.walkers[m].allpars
            mp = len( crospar )
            if mp == 0:
                continue
            if np < mp:
                mp = np

            # take parameters randomly from param and crospar
#            perm = self.rng.permutation( fitIndex )
#            kp = 1 + self.rng.randint( mp - 2 )         # at least 1 par of each walker
#            perm = perm[:kp]
#            param[perm] = crospar[perm]

            # mix parameters randomly
            f = self.rng.rand( nf )
            param[fitIndex] = f * param[fitIndex] + ( 1 - f ) * crospar[fitIndex]

            Ltry = self.errdis.logLikelihood( model, param )
            if Ltry >= lowLhood:
                self.reportSuccess( )
                self.setSample( walker, model, param, Ltry )
                return nf
            elif kk <= self.maxtrials :
                param = walker.allpars[:]
                self.reportReject( )
            else :
                self.reportFailed()
                return 0

        return 0





import numpy as numpy
import scipy
import math

from .ScaledErrorDistribution import ScaledErrorDistribution
from .NoiseScale import NoiseScale


__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester


class CauchyErrorDistribution( ScaledErrorDistribution ):
    """
    To calculate a Cauchy or Lorentz likelihood.
    ..math ::
        f( x ) = s / ( \pi * ( s^2 + x^2 ) )

    where x = residual and s = scale

    The function is mostly used to calculate the likelihood L, or easier
    to use log likelihood, logL.
    .. math::
        logL = N ( \log( s ) - \log( \pi ) ) - \sum( \log( x^2 + s^2 ) )

    The use of weights is not possible in this error distribution.

    s is a hyperparameter, which might be estimated from the data.

    Author       Do Kester.

    """

    LOGPI = math.log( math.pi )

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, xdata, data, weights=None, scale=1.0, limits=None,
                  copy=None ):
        """
        Default Constructor.

        Parameters
        ----------
        xdata : array_like
            input data fro the model (independent data)
        data : array_like
            data to be fitted (dependent data)
        weights : array_like
            weights to be used (no weights are possible in Cauchy)
        scale : float
            noise scale
        limits : None or list of 2 floats [low,high]
            None : no limits implying fixed scale
            low     low limit on scale (needs to be >0)
            high    high limit on scale
            when limits are set, the scale is *not* fixed.
         copy : CauchyErrorDistribution
            distribution to be copied.

        """
        if weights is not None:
            raise ValueError( "Weights are not possible in Cauchy distributions" )

        super( CauchyErrorDistribution, self ).__init__( xdata, data,
                scale=scale, limits=limits, copy=copy )

    def copy( self ):
        """ Return copy of this.  """
        return CauchyErrorDistribution( self.xdata, self.data, copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes.

        """
        if name == "res2" :
            object.__setattr__( self, name, value )
        else :
            super( CauchyErrorDistribution, self ).__setattr__( name, value )

    def acceptWeight( self ):
        """
        True if the distribution accepts weights.
        False for this distribution.
        """
        return False

    def getScale( self, model ) :
        """
        Return the noise scale as calculated from the residuals.

        Parameters
        ----------
        model : Model
            the model involved
        """
        res = self.getResiduals( model )
        self.res2 = res * res

        scale = scipy.optimize.bisect( self.funct, 0.001, 100.0 )
        return scale

    def funct( self, scale ) :
        return ( numpy.sum( numpy.log( self.res2 + scale * scale ) ) -
                self.sumweight * ( 2 * math.log( scale ) + math.sqrt( 2.0 ) ) )

    #  *********LIKELIHOODS***************************************************
    def logLikelihood( self, model, allpars ):
        """
        Return the log( likelihood ) for a Cauchy distribution.
        Cauchy distr : f( x ) = s / ( pi * ( s^2 + x^2 ) )

        where x = residual and s = scale

        Parameters
        ----------
        model : Model
            model to calculate mock data
        allpars : array_like
            parameters of the problem

        """
        self.ncalls += 1
        np = model.npchain
        scale = allpars[np]
        res2 = numpy.square( self.getResiduals( model, allpars[:np] ) )
        return ( self.ndata * ( math.log( scale ) - self.LOGPI ) -
                 numpy.sum( numpy.log( res2 + scale * scale ) ) )

    def partialLogL( self, model, allpars, fitIndex ) :
        """
        Return the partial derivative of log( likelihood ) to the parameters
        in fitIndex.

        Parameters
        ----------
        model : Model
            model to calculate mock data
        allpars : array_like
            parameters of the problem
        fitIndex : array_like
            indices of parameters to be fitted

        """
        self.nparts += 1
        np = model.npchain
        scale = allpars[np]
        res = self.getResiduals( model, allpars[:np] )
        r2s = res * res + scale * scale
        dM = model.partial( self.xdata, allpars[:np] )

        dL = numpy.zeros( len( fitIndex ), dtype=float )
        i = 0
        for k in fitIndex :
            if k < np :
                dL[i] = 2 * numpy.sum( res * dM[:,k] / r2s )
            else :
                dL[i] = self.ndata / scale - numpy.sum( 2 * scale / r2s )
            i += 1
        return dL

    def __str__( self ) :
        return "Cauchy error distribution"

